from pyglet import gl
from pyleap.shape.circle import Shape

class Point(Shape):
    """ 点 Point """

    def __init__(self, x=100, y=100, color="orange", point_size=1):
        super().__init__(x, y, color, gl=gl.GL_POINTS)

        self.point_size = point_size # only for point

    def update_points(self):
        self.points = (self.x, self.y)



from pyglet import gl
from pyleap.shape.shape import Shape
import math

class Polygon(Shape):
    """
    多边形：Polygon
    """
    def __init__(self, *args):

        """ 多边形

            颜色： color 默认为 "orange"
        """

        if isinstance(args[-1], str):
            color = args[-1]
            points = args[:len(args)-1]

        else:
            color = "orange"
            points = args

        x = sum(points[::2]) / (len(points) / 2)
        y = sum(points[1::2]) / (len(points) / 2)

        super().__init__(x, y, color, gl=gl.GL_POLYGON)

        self.points = points


"""基本图形：椭圆

椭圆是以圆心为中心的多边形（72边形）的形式进行拟合
"""


from pyglet import gl
from math import sin, cos, pi
from pyleap.shape.shape import Shape


class Ellipse(Shape):
    """ 基本图形：椭圆 Ellipse 
    """

    def __init__(self, x=100, y=100, r_x=50, r_y=30, color="orange"):
        """ 
        圆心： x、y,        默认为100, 100
        半径： r_x, r_y    默认为50, 30
        颜色： color,      默认为 "orange"
        """
        super().__init__(x, y, color, gl=gl.GL_POLYGON)
        self.r_x = r_x
        self.r_y = r_y

    def update_points(self):
        """ 椭圆的近似图形：72边形 """
        n = 72
        d = pi * 2 / n
        x, y, r_x, r_y = self.x, self.y, self.r_x, self.r_y

        ps = []
        for i in range(n):
            ps += [(x + r_x * sin(d * i)), (y + r_y * cos(d * i))]
        self.points = tuple(ps)



from pyglet.window import mouse as default_mouse
from pyleap import null


class Mouse():
    """ 鼠标类
    属性
    x: x坐标
    y: y坐标

    LEFT: bool, 左键是否被按下
    RIGHT： bool, 右键是否被按下

    事件
    on_move(f): 鼠标移动时
    on_press(f): 左键点击时
    on_release(f): 左键松开时
    on_right_press(f): 右键点击时
    on_right_release(f): 右键松开时

    """

    def __init__(self):
        self.x = 0
        self.y = 0
        self._move = null

        self._press = null
        self._release = null

        self._right_press = null
        self._right_release = null

        self.LEFT = False
        self.RIGHT = False

    def on_move(self, func):
        """ 注册鼠标移动事件 """
        self._move = func

    def on_press(self, func):
        """ 注册鼠标点击事件 """
        self._press = func

    def on_release(self, func):
        """ 注册鼠标点击事件 """
        self._release = func

    def on_right_press(self, f):
        """ 注册鼠标右键点击事件 """
        self._right_press = f

    def on_right_release(self, f):
        """ 注册鼠标右键点击事件 """
        self._right_release = f
        
    def move(self):
        """ 触发鼠标移动事件 """
        self._move()

    def press(self):
        """ 触发鼠标左键点击事件 """
        self.LEFT = True
        self._press()

    def release(self):
        """ 触发鼠标左键松开事件 """
        self.LEFT = False
        self._release()

    def right_press(self):
        """ 触发鼠标右击事件 """
        self.RIGHT = True
        self._right_press()

    def right_release(self):
        """ 触发鼠标右键松开事件 """
        self.RIGHT = False
        self._right_release()

mouse = Mouse()



""" 基本图形： 线段

包含2个点

属性：
points： （x1, y1, x2, y2）
"""
from pyleap.shape.shape import Shape
from pyglet import gl
import math


class Line(Shape):
    def __init__(self, x1=100, y1=100, x2=200, y2=200, line_width=1, color="orange"):
        """ 线段
        线段宽度： line_width， 默认为1
        颜色： color, 默认为"orange"
        """
        x = (x1 + x2) / 2
        y = (y1 + y2) / 2
        super().__init__(x, y, color, gl=gl.GL_LINES, line_width=line_width)
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2


    def update_points(self):
        self.points = (self.x1, self.y1, self.x2, self.y2)




"""
判断2个基本图形是否发生碰撞，

如果没有碰撞，则返回False
如果碰撞，则返回其中一个碰撞的坐标点，如 {'x': 100, 'y': 200}

"""
from pyleap.mouse import mouse


__all__ = ['shape_clicked', 'CollisionMixin' ]

class P:
    def __init__(self, x, y):
        self.x = x
        self.y = y

def shape_clicked(shape):
    shape.transform.update_points(shape.points)
    return point_in_points((mouse.x, mouse.y), shape.transform)

def point_in_points(p, s2):
    x = p[0]
    y = p[1]
    cross_count = 0
    for i in range(0, len(s2.points), 2):
        x1 = s2.points[i-2]
        y1 = s2.points[i-1]
        x2 = s2.points[i]
        y2 = s2.points[i+1]
        if (x1 < x and x2 < x) or (y1-y)*(y2-y) > 0 or (y1==y2) or (y2==y):
            continue
        cross_x = x1 - (x1-x2)*(y1-y)/(y1-y2)
        if(cross_x >= x):
            cross_count += 1

    return cross_count % 2



def points_in_points(s1, s2):
    for i in range(0, len(s1.points), 2):
        x = s1.points[i]
        y = s1.points[i+1]

        # outside the collide rect
        if(x < s2.min_x or x > s2.max_x or y < s2.min_y or y > s2.max_y):
            continue

        if(point_in_points((x, y), s2)):
            return P(x, y)

    return False

def lines_cross(s1, s2):
    for i in range(0, len(s1.points), 2):
        x1 = s1.points[i-2]
        y1 = s1.points[i-1]
        x2 = s1.points[i]
        y2 = s1.points[i+1]
        for i in range(0, len(s2.points), 2):
            x3 = s2.points[i-2]
            y3 = s2.points[i-1]
            x4 = s2.points[i]
            y4 = s2.points[i+1]
            p = line_cross(x1, y1, x2, y2, x3, y3, x4, y4)
            if p: return p

    return False

def line_cross(x1, y1, x2, y2, x3, y3, x4, y4):
    """ 判断两条线段是否交叉 """
    # out of the rect
    if min(x1, x2) > max(x3, x4) or max(x1, x2) < min(x3, x4) or \
       min(y1, y2) > max(y3, y4) or max(y1, y2) < min(y3, y4):
        return False

    # same slope rate
    if ((y1 - y2) * (x3 - x4) == (x1 - x2) * (y3 - y4)):
        return False

    if cross_product(x3, y3, x2, y2, x4, y4) * cross_product(x3, y3, x4, y4, x1, y1) < 0 or \
       cross_product(x1, y1, x4, y4, x2, y2) * cross_product(x1, y1, x2, y2, x3, y3) < 0:
        return False

    # get collide point
    b1 = (y2 - y1) * x1 + (x1 - x2) * y1
    b2 = (y4 - y3) * x3 + (x3 - x4) * y3
    D = (x2 - x1) * (y4 - y3) - (x4 - x3) * (y2 - y1)
    D1 = b2 * (x2 - x1) - b1 * (x4 - x3)
    D2 = b2 * (y2 - y1) - b1 * (y4 - y3)

    return P(D1 / D, D2 / D)


def cross_product (x1, y1, x2, y2, x3, y3):
    """ 叉乘
    vector 1: x1, y1, x2, y2
    vector 2: x1, y1, x3, y3
    """
    return (x2 - x1) * (y3 - y1) - (x3 - x1) * (y2 - y1)


class CollisionMixin():
    """ 碰撞图形
    将任意图形转换为多边形进行判断，图形需要拥有points属性

    按照以下步骤进行判断
    1. 如果图形的外接矩形没有碰撞，返回False
    2. 如果图形的点在目标图形内，返回这个点
    3. 如果图形的边和目标图形的边相交，返回相交点
    4. 返回False

    """

    def update_collision_rect(self):
        """ 获取外接矩形 """
        self.min_x = min(self.points[::2])
        self.max_x = max(self.points[::2])
        self.min_y = min(self.points[1::2])
        self.max_y = max(self.points[1::2])

    def collide(self, s2):
        """ 判断图形是否碰到了另外一个图形 """
        s1 = self
        if not (s1.points and s2.points):
            return False

        t1 = s1.transform
        t2 = s2.transform

        t1.update_points(s1.points)
        t2.update_points(s2.points)

        # 更新外接矩形
        t1.update_collision_rect()
        t2.update_collision_rect()

        # simple collide rect
        if not (t1.min_x < t2.max_x and t1.max_x > t2.min_x \
                and t1.min_y < t2.max_y and t1.max_y > t2.min_y):
            return False

        return points_in_points(t1, t2) or \
               points_in_points(t2, t1) or \
               lines_cross(t1, t2)

    def on_press(self, func):
        """ 当图形被点击时，触发func函数 """
        self._press = func




""" 资源加载扩展
当资源是一个链接时，将先下载资源保存到本地，再进入游戏

"""

import os
import hashlib
import re

import ssl
import urllib.request

ctx = ssl.create_default_context()
ctx.check_hostname = False
ctx.verify_mode = ssl.CERT_NONE


class Resource():

    def __init__(self, path="download"):
        """ 设置保存路径 """
        self.path = path

    def get(self, url):
        if not self.is_url(url):
            return url

        filename = self.md5_8_name(url)
        fullname =  '{}{}{}'.format(self.path, os.sep, filename)

        if not os.path.exists(fullname):
            self.download(url, fullname)

        return fullname

    def md5_8_name(self, url):
        """ 把下载的文件重命名为地址的md5前8位 """
        m = hashlib.md5()
        m.update(url.encode('utf-8'))
        return m.hexdigest()[:8] + os.path.splitext(url)[1]

    def download(self, url, fullname):
        if not os.path.exists(self.path):
            os.makedirs(self.path)

        print("Downloading: " + url)
        # urllib.request.urlretrieve(url, filename=fullname)
        with urllib.request.urlopen(url, context=ctx) as u, \
            open(fullname, 'wb') as f:
            f.write(u.read())

    def is_url(self, url):
        regex = re.compile(
                r'^(?:http|ftp)s?://' # http:// or https://
                r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|' #domain...
                r'localhost|' #localhost...
                r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})' # ...or ip
                r'(?::\d+)?' # optional port
                r'(?:/?|[/?]\S+)$', re.IGNORECASE)
        return re.match(regex, url) is not None



from pyleap.window import window
from pyglet.window import key as default_key

from pyleap import null

class SingleKey():

    def __init__(self):
        self._press = null
        self._release = null
        self.pressed = False
    
    def on_press(self, f):
        self._press = f

    def press(self):
        self.pressed = True;
        self._press()

    def on_release(self, f):
        self._release = f

    def release(self):
        self.pressed = False;
        self._release()


class Key(dict):
    """ 
    当按下A键时，key.A.pressed 为True，否则为False

    键名

    字母：A, B, C, D, E, F, G...
    数字：因为数字不能直接作为变量, 变为_1, _2, ... 如 key._1
    方向键：LEFT，UP，RIGHT，DOWN
    功能键：HOME，PAGEUP，PAGEDOWN，END，BEGIN，BACKSPACE，TAB，LINEFEED
        CLEAR，RETURN，ENTER，PAUSE，SCROLLLOCK，SYSREQ，ESCAPE，SPACE
    功能键：F1, F2, F3..
    切换键：LSHIFT, RSHIFT, LCTRL, RCTRL，CAPSLOCK，LMETA，RMETA，LALT，RALT
        LWINDOWS， RWINDOWS，LCOMMAND，RCOMMAND，LOPTION，ROPTION
    数字区：NUM_0，NUM_1，NUM_2...


    """

    def __init__(self, **kw):
        super().__init__(**kw)

        for symbol, name in default_key.__dict__['_key_names'].items():
             s_k = SingleKey()
             self[symbol] = s_k
             self[name] = s_k

    def __getattr__(self, key):
        return self[key.upper()]



key = Key()


__all__ = ['key']


import pyglet
import sys

from pyleap.resource import Resource

# 用于管理资源
rss = Resource()

def null():
    """ 空函数 """
    pass

stop = pyglet.clock.unschedule

def repeat(f, dt=1/60):
    """ 重复执行函数f，时间间隔dt """
    stop(f)
    pyglet.clock.schedule_interval(f, dt)


def run_after(f, dt=1/60):
    """ 在一定时间之后执行 """
    pyglet.clock.schedule_once(f, dt)


def run():
    """ 在程序的最后运行，进入循环阻塞 """
    pyglet.app.run()
    

"""all_shapes, use for track all the shapes draw on the windows
"""
all_shapes = set()

__all__ = ['null', 'run', 'stop', 'repeat', 'run_after', 'all_shapes', 'rss']



from pyglet import gl
from pyleap.shape.shape import Shape


class Triangle(Shape):
    """ Triangle """

    def __init__(self, x1=100, y1=100, x2=200, y2=150, x3=200, y3=100,
                 color='orange'):
        x = (x1 + x2 + x3) / 3
        y = (y1 + y2 + y3) / 3
        super().__init__(x, y, color, gl=gl.GL_TRIANGLES)
        self.points = (x1, y1, x2, y2, x3, y3)





""" 音效对象
支持的音效格式有（不限于以下列出的格式）：

* AU
* MP2
* MP3
* OGG/Vorbis
* WAV
* WMA

在项目中使用音效，除wav格式的音效外，其他的音效资源需要安装Avbin才能播放。 Avbin下载地址：http://avbin.github.io/AVbin/Download.html

可以使用网络地址链接，如果使用的是网络地址链接，则会下载该资源到本地，下次运行时将直接读取本地资源



"""

import pyglet
from pyleap.util import rss


class Audio(pyglet.media.Player):
    """ 音效对象
   
    Audio(url, loop=False)

    属性

    time    ： 当前的播放时间（秒），只读属性，通过seek()方法更改时间
    playing  : bool值，当前是否在播放，只读属性
    volume  ： float值，音量大小，从0到1，
    loop     : bool值，是否重复播放

    方法

    play() : 开始或者继续播放当前音效
    pause() : 暂停播放当前音效
    seek()  : 设置播放的时间

    """

    def __init__(self, src, loop=False):
        """ Decoding sounds can be processor-intensive and may introduce latency, particularly for short sounds that must be played quickly, such as bullets or explosions. You can force such sounds to be decoded and retained in memory rather than streamed from disk by wrapping the source in a StaticSource:"""
        super().__init__()
        _source = pyglet.media.load(rss.get(src))
        self._source_group = pyglet.media.SourceGroup(_source.audio_format, None)
        self._source_group.loop = loop
        self._source_group.queue(_source)
        self.queue(self._source_group)

    @property
    def loop(self):
        return self._source_group.loop
    
    @loop.setter
    def loop(self, loop):
        self._source_group.loop = loop
    




""" 基本图形类 Shape

所有图形都继承与Shape，Shape拥有以下两个主要的属性
1. Collision，主要用于检测碰撞以及点击事件
2. Transform，用于图形的变换处理

"""

import pyglet
from pyglet import gl

from pyleap.transform import Transform, TransformMixin
from pyleap.collision import CollisionMixin
from pyleap.color import color_to_tuple
from pyleap.util import all_shapes


class Shape(CollisionMixin, TransformMixin):
    """ base shape class """

    def __init__(self, x, y, color="orange", gl=gl.GL_LINE_LOOP,
                 line_width=1):
        """ 默认参数
        颜色 color: "orange"
        线条粗细 line_width： 1
        """
        self.x = x
        self.y = y
        self.color = color
        self.gl = gl
        self.transform = Transform()

        # 图形的线条宽度，默认为1
        self.line_width = line_width

        # 图形的近似多边形 (x1, y1, x2, y2, ...)
        self.points = ()

        # 用于记录press事件函数
        self._press = None

        # 仅Point类point_size属性有效
        self.point_size = 1

        # 透明度 opacity
        self.opacity = 1

    def draw(self):
        """ 使用draw方法将图形绘制在窗口里 """
        self.update_all()
        self.vertex_list.draw(self.gl)

    def stroke(self):
        """ 使用stroke方法将图形绘制在窗口里，仅对基本的几何图形有效 """
        self.update_all()
        self.vertex_list.draw(gl.GL_LINE_LOOP)

    def update_all(self):
        """ 在绘制之前，针对形变进行计算，通过设置openGL的属性来达到绘制出变形的图形 """
        self.update_points()
        self.update_vertex_list()
        self.update_anchor()

        gl.glLoadIdentity() # reset gl
        gl.glLineWidth(self.line_width)
        gl.glPointSize(self.point_size)
        self.transform.update_gl()

        # handle shapes click envets
        all_shapes.discard(self)
        if(self._press != None):
            all_shapes.add(self)     


    def update_points(self):
        """ translate shapes to points，在子类中实现 """
        pass

    def update_vertex_list(self):
        """ 使用pyglet来绘制基本图形之前，转为pyglet识别的属性 """
        color = color_to_tuple(self.color, self.opacity)
        length = len(self.points) // 2
        self.vertex_list = pyglet.graphics.vertex_list(
            length,
            ('v2f', self.points),
            ('c4B', color * length))

    def update_anchor(self):
        """ 如果是使用set_anchor_rate来设定锚点，那么就需要不停的更新锚点的位置 """
        t = self.transform
        self.update_collision_rect()
        if t.anchor_x_r and t.anchor_y_r:
            t.anchor_x = self.min_x + (self.max_x - self.min_x) * t.anchor_x_r
            t.anchor_y = self.min_y + (self.max_y - self.min_y) * t.anchor_y_r

    


import pyglet
from pyleap.window import window
from pyleap.color import color_to_tuple
from pyleap.util import all_shapes

from pyleap.shape.rectangle import Rectangle


class Text(Rectangle):
    """ Text
    src: 文本内容
    font_size: 文字高度
    """

    def __init__(self, src, x=window.center_x, y=window.center_y,
                 font_size=16, color="orange"):
        """ """

        self.text = pyglet.text.Label(text=src, font_size=font_size)
        super().__init__(x=x, y=y, h=font_size, color=color)
        self.update_width()

    def draw(self):
        self.update_all()
        self.text.x = self.x
        self.text.y = self.y
        self.text.font_size = self.h
        self.text.color = color_to_tuple(self.color)
        self.text.draw()

    @property
    def src(self):
        return self.text.text

    @src.setter
    def src(self, src):
        self.text.text = src
        self.update_width()

    @property
    def font_size(self):
        return self.h

    @font_size.setter
    def font_size(self, font_size):
        self.h = font_size
        self.update_width()

    def update_width(self):
        self.w = self.h * len(self.src) * 0.618
    


from pyleap.window import window
from pyleap.mouse import mouse
from pyleap.collision import shape_clicked
from pyleap.util import all_shapes
from pyleap.key import key


@window.event
def on_mouse_motion(x, y, dx, dy):
    """ 当鼠标没有按下时移动的时候触发 """
    mouse.x, mouse.y = x, y
    mouse.move()
    window.update_caption(mouse)


@window.event
def on_mouse_drag(x, y, dx, dy, buttons, modifiers):
    """ 当鼠标按下并且移动的时候触发 """
    mouse.x, mouse.y = x, y
    mouse.move()


class MouseKeyCode:
    """
    button
    1 - 左键
    2 - 滚轮
    4 - 右键
    """
    LEFT = 1
    MIDDLE = 2
    RIGHT = 4

@window.event
def on_mouse_press(x, y, button, modifiers):
    """ 按下鼠标时 

    """ 
    if button == MouseKeyCode.LEFT:
        mouse.press()
    elif button == MouseKeyCode.RIGHT:
        mouse.right_press()

    # 判断是否有图形的点击事件被触发了
    shapes = list(all_shapes)
    while shapes:
        shape = shapes.pop()
        if(shape._press and shape_clicked(shape)):
            shape._press()


@window.event
def on_mouse_release(x, y, button, modifiers):
    """ 松开鼠标时 """ 
    if button == MouseKeyCode.LEFT:
        mouse.release()
    elif button == MouseKeyCode.RIGHT:
        mouse.right_release()


@window.event
def on_key_press(symbol, modifiers):
    """ 当键盘按键按下时触发 """
    try:
        key[symbol].press()
    except:
        pass


@window.event
def on_key_release(symbol, modifiers):
    """ 当键盘按键松开时触发 """
    try:
        key[symbol].release()
    except:
        pass
        


import pyglet
from pyglet import gl
from pyleap.util import all_shapes
import platform

# disable debug gl option 
pyglet.options['debug_gl'] = False

class Window(pyglet.window.Window):
    """ 
    属性 Attributes

    caption
    str，窗口标题，只读.

    width/w
    窗口宽度，单位像素点，只读。

    height/h
    窗口高度，单位像素点，只读。

    方法 Methods

    clear()
    清除窗口里的内容

    set_caption(caption)
    设置窗口标题
    Parameters: caption (str or unicode)

    set_size(width, height)
    设置窗口大小
    """

    def __init__(self):
        """ 初始化，创建一个窗口 """

        sysstr = platform.system()
        plat = pyglet.window.get_platform()
        display = plat.get_default_display()
        screen = display.get_default_screen()

        template = gl.Config(
            alpha_size=8,
            sample_buffers=1,       # 抗锯齿设置
            samples=4
            )
            
        if(sysstr =="Windows"):
            try:
                configs = screen.get_matching_configs(template)
                if not configs:
                    raise("Error")
                else:
                    super().__init__(config=configs[0])
            except:
                print("你的显卡不支持高级特性，部分功能将无法使用")
                super().__init__()
        else: # Mac or Linux
            super().__init__()
        
        self.set_caption("LeapLearner")

    @property
    def w(self):
        """ width """
        return self.width

    @property
    def h(self):
        """ height """
        return self.height

    @property
    def center_x(self):
        return self.width // 2

    @property
    def center_y(self):
        return self.height // 2

    @property
    def caption(self):
        return self._title
    
    def set_caption(self, caption):
        self._title = caption

    def update_caption(self, mouse):
        """ 添加坐标显示 """
        caption = "{}  x: {}, y: {}".format(self._title, mouse.x, mouse.y)
        super().set_caption(caption)

    def clear(self):
        all_shapes.clear()
        super().clear()

    def show_axis(self):
        from pyleap import Line, Text
        for x in range(0, self.w, 100):
            Line(x, 0, x, self.h, 1, '#eeaa00').draw()
            Text(str(x), x+2, 2, 10).draw()
        for y in range(0, self.h, 100):
            Line(0, y, self.w, y, 1, '#eeaa00').draw()
            Text(str(y), 2, y+2, 10).draw()


window = Window()

# 必须放在window后面

# 抗锯齿
gl.glEnable(gl.GL_LINE_SMOOTH);
gl.glHint(gl.GL_LINE_SMOOTH_HINT, gl.GL_DONT_CARE);

gl.glEnable(gl.GL_POLYGON_SMOOTH);
gl.glHint(gl.GL_POLYGON_SMOOTH_HINT, gl.GL_DONT_CARE);

# 抗锯齿-多样本缓冲(Multisample Buffer)
gl.glEnable(gl.GL_MULTISAMPLE);

# 支持透明
gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA);
gl.glEnable(gl.GL_BLEND);

